@@ -9,8 +9,7 @@ class ViewAuto(base.View):
         # TODO: The auto view has little justification now that views implement render_priority,
         # but we keep it around for now to not touch more parts.
         priority, view = max(
-            (v.render_priority(data, **metadata), v)
-            for v in contentviews.views
+            (v.render_priority(data, **metadata), v) for v in contentviews.views
         )
         if priority == 0 and not data:
             return "No content", []
